# Processor.py
'''To be used for handling any operational component of wheat rust early warning
system. 

Command-line options specify which component and which json configuration 
files to run, and complete any remaining elements of the configuration files.

Component varies by choices of process_pre_job(), process_in_job() and 
process_EWS_plotting().

This runs all dates of a single forecast.

Example usage::
    $ conda activate py3EWSepi
    $ python Processor.py -p Deposition --islive --config config_Nepal_template.json --noupload -s 20200715
or::
    $ ./run_Processor.sh -p Deposition --islive --config config_Nepal_template.json --noupload -s 20200715
'''
from abc import abstractmethod, ABCMeta
from typing import List

from ews.coordinator.utils import processor_utils
from ews.coordinator.utils.jobstatus import Jobstatus
from ews.coordinator.utils.processor_utils import end_script, end_job, append_item_to_list, \
    clear_up, parse_json_file_with_tokens


print("Make sure to `conda activate py3EWSepi` environment!")
print("Make sure that flagdir package is available (on PYTHONPATH)")

import argparse
import datetime
import json
import logging
import os
from pathlib import Path
import sys


# submodules of this project

logger = logging.getLogger(__name__)


class ProcessorBase:

    __metaclass__ = ABCMeta

    def __init__(self) -> None:
        super().__init__()
        time_now = datetime.datetime.today()
        self.todayString = time_now.strftime('%Y%m%d')
        self.nowString = time_now.strftime('%Y%m%d-%H%M-%S')

    def parse_and_check_args(self) -> dict:

        # define the command line arguments
        my_parser = argparse.ArgumentParser(
            description = 'Command-line arguments for coordinator script of env suitability model')

        # Add the arguments
        # positional arguments do not start with - or -- and are always required
        # optional arguments start with - or -- and default is required = False

        my_parser.add_argument(
                '-sc', '--sys_config',
                metavar = 'path',
                type = str,
                dest = 'sys_config_path',
                required = True,
                help = '''path to the system config file. This file contains values which are common to all Processor 
                components''')

        my_parser.add_argument(
                '-c', '--config',
                metavar = 'path',
                type = str,
                nargs = '+',  # require at least one path
                dest = 'config_paths',
                required = True,
                help = '''path to a config file(s). More than one can be provided,
        in which case each is worked on in turn (e.g. one each for stripe, stem, leaf).
        Do not place other options between these.''')

        my_parser.add_argument(
                '-l', '--loglevel',
                action = 'store',
                choices = list(["debug", "info", "warning", "error", "critical"]),
                default = 'info',
                help = 'verbosity of log messaging (debug, info, warning, error, critical)\n default is debug',
                dest = 'log_level',  # this names the attribute that will be parsed
        )

        my_parser.add_argument(
                '--islive',
                action = 'store_true',
                help = 'If live, email messages are sent to maintainers for warning and errors',
                dest = 'live',
        )

        my_parser.add_argument(
                '-s', '--start-date', '-i', '--initial-date',
                metavar = 'YYYYMMDD',
                action = 'store',
                default = self.todayString,
                help = 'Initial day of calculation, starting at 00 UTC (Default is today)',
                dest = 'start_date',
        )

        my_parser.add_argument(
                '--noupload',
                action = 'store_true',
                help = 'whether results of script should be saved to willow public directory'
        )

        my_parser.add_argument(
                '--clearup',
                action = 'store_true',
                help = 'whether to delete mid-process files at the end of a successful job',
                dest = 'clearup',
        )

        # get an object holding all of the args
        args = my_parser.parse_args()

        # Check the args

        print(f"Command-line options are:\n{args}")

        if not isinstance(args.config_paths, list):
            print('Expecting a list of config paths')
            raise RuntimeError

        # check the startstring
        if args.start_date is not self.todayString:
            try:
                # check date string is formatted correctly
                provided_start_date = datetime.datetime.strptime(args.start_date, '%Y%m%d')
                today_date = datetime.datetime.strptime(self.todayString, '%Y%m%d')

                # early limit is quite arbitrary, but this is earliest year of available survey data for Ethiopia
                date_limit_early = datetime.datetime.strptime('20070101', '%Y%m%d')
                assert date_limit_early < provided_start_date
                assert provided_start_date <= today_date

            except (ValueError, AssertionError) as e:
                raise Exception(
                    "Provided start date string is formatted incorrectly or out of range, or end date not also defined")

        dictionary: dict = vars(args)
        return dictionary


    # def build_sys_config(self, sys_config_path: str) -> dict:
    #     '''This config obtains aspects of each dict in configs that must be common to
    #     them all. '''
    #
    #     # initialise universal config
    #     # sys_config = {
    #     #     "ProjectRoot":      set(),
    #     #     'WorkspacePathout': set(),
    #     #     "MetoFTP":          set(),
    #     #     "WorkspacePath":    set(),
    #     #     "ResourcesPath":    set()
    #     # }
    #
    #     sys_config = {}
    #     # keys = sys_config.keys()
    #
    #     # get value of each key from each config file
    #     try:
    #         sys_config = open_and_check_config(sys_config_path)
    #         # for key in keys:
    #         #     if key in config_i.keys():
    #         #         sys_config[key].add(config_i[key])
    #     except:
    #         logger.exception(f"Failure in opening or checking config {sys_config_path}")
    #         end_script(premature = True)
    #
    #     # Check for and keep only one value per key
    #     # for key in keys:
    #     #     if len(sys_config[key]) > 1:
    #     #         logger.error(f"Config files point to multiple {key} but this script can only handle one.")
    #     #         end_script(premature = True)
    #     #
    #     #     sys_config[key] = sys_config[key].pop()
    #
    #     # config_tokens = {'ProjectRoot':      sys_config['ProjectRoot'],
    #     #                  'WorkspacePathout': sys_config['WorkspacePathout'],
    #     #                  'MetoFTP':          sys_config['MetoFTP'],
    #     #                  'WorkspacePath':    sys_config['WorkspacePath'],
    #     #                  'ResourcesPath':    sys_config['ResourcesPath']}
    #
    #     return sys_config

    def prepare_job_directory(self, job_path):
        """
        create job directory or archive if already exists (due to a rerun)

        :param job_path:
        :return:
        """
        if os.path.exists(job_path):
            archived_dir_path = f"{job_path}-ARCHIVE_{self.nowString}"
            message = f"Job path {job_path} already exists, so moving it to {archived_dir_path}"
            print(message)
            os.rename(job_path, archived_dir_path)

        Path(job_path).mkdir(parents = True, exist_ok = False)


    def run_process(self,
                    args: dict) -> None:

        # check initial state of each config file, and gather terms that must apply
        # across all provided configs

        sys_config_path: str = args['sys_config_path']
        config_paths: List[str] = args['config_paths']
        component: str = args['component']
        short_name: str = args['short_name']
        start_date: str = args['start_date']
        clearup: bool = args['clearup']

        # load universal configuration
        sys_config = parse_json_file_with_tokens(sys_config_path)
        sys_config['StartString'] = start_date

        # determine job directory
        workspace_path = sys_config['WorkspacePathout']
        job_path: str = f'{workspace_path}{short_name}_{start_date}'
        self.prepare_job_directory(job_path)

        is_live: bool = args["live"]
        log_file_path = f"{job_path}/log.txt"

        """
        Now we have the job_dir defined, we can set up the logging
        """
        if 'LoggingConfigPath' not in sys_config.keys():
            raise Exception(f"'LoggingConfigPath' not found in sys_config. Either use the default template file in "
                            f"configs/logger/template_log_config.json, or copy the template away and edit")

        log_config_path = sys_config['LoggingConfigPath']
        processor_utils.setup_logging(log_file_path,
                                      log_config_path,
                                      is_live,
                                      args['log_level'])

        logger.info("==========")
        logger.info(f"Logging started at {datetime.datetime.now().strftime('%Y %b %d %H:%M:%S')}")
        logger.info(f"Universal config is\n{json.dumps(sys_config, indent = 2)}")
        logger.info(f"Job path will be {job_path}")
        # workspace_path = sys_config['WorkspacePathout']

        # note that file date represents preceding 3 hours, so day's data starts at file timestamp 0300 UTC
        start_time: datetime = datetime.datetime.strptime(start_date + '03', '%Y%m%d%H')
        start_time_string: str = start_time.strftime('%Y-%m-%d-%H%M')

        # run any checks before creating a job directory
        # if this fails, then make a note once there is a job directory
        ready = self.process_pre_job(args)

        # lock job directory
        status: Jobstatus
        with Jobstatus(job_path) as status:

            # check for a status file in job directory
            if status.had_initial_status:
                logger.info(f"Job path already exists and has status {status.status}")

                end_script(premature = status.status not in ['SUCCESS', 'INPROGRESS'])

            logger.info(f"Current status of job directory is {status.status}")

            # make a record if process_pre_job failed
            if not ready:
                logger.error(
                    f"Process_pre_job raised an error so making a record in the job file. For details, see earlier warnings log")
                status.reset('ERROR')

                end_job(status, ignore_inprogress = True, premature = False)

            # files and directories that will be earmarked for removal after a
            # successful job
            paths_to_clear = []

            logger.info('Starting to work on each configuration')

            for configIndex, configtemplate in enumerate(config_paths):

                config_paths_length: int = len(config_paths)
                logger.info(f'Working on config {configIndex + 1} of {config_paths_length}')

                try:
                    configjson: dict = parse_json_file_with_tokens(configtemplate, sys_config)
                    # then add the sys_config keys and values to the configjson
                    for k, v in sys_config.items():
                        if k not in configjson.keys():
                            configjson[k] = v
                        else:
                            logger.warning(f"Key {k} already present in run config - not adding key with same name from "
                                           f"the sys_config")

                except:
                    logger.exception(f"Failure in opening or checking config {configtemplate}")
                    # TODO: This case should test flagdir.jobStatus.__exit__()
                    raise  # endJob('ERROR',premature=True)

                # provide specific case details to template config

                configjson['StartTime'] = start_time_string
                configjson['StartString'] = start_date

                # from configtemplate create configFileName to describe the specific job
                config_file_name = f"{os.path.basename(configtemplate).replace('.json', '')}_{component}"

                configjson['ConfigFilePath'] = config_file_name

                # write the complete configuration file to job directory
                with open(f"{job_path}/{config_file_name}.json", 'w') as write_file:
                    json.dump(configjson, write_file, indent = 4)

                # proc_description = universal_config['ProcessInJob']
                proc_description = 'ProcessInJob'
                try:
                    proc_out: dict = self.process_in_job(job_path, status, configjson, component)
                except:
                    logger.exception(f"Error in process_in_job")
                    status.reset('ERROR')
                    end_job(status, premature = True)

                # Set default case
                # This would be improved by implementing a class structure
                if proc_out is None:
                    proc_out = {
                        'output':  None,
                        'clearup': None}

                if 'clearup' in proc_out.keys():
                    append_item_to_list(
                            proc_out['clearup'],
                            paths_to_clear,
                            proc_description,
                            status)

                # Run EWS-plotting command

                proc_description = 'ProcessEWSPlotting'
                try:
                    self.process_post_job(job_path, configjson)
                except:
                    logger.exception(f"Error in {proc_description}()")
                    status.reset('ERROR')
                    end_job(status, premature = True)

                logger.info('Finished with EWS-Plotting, appending images to list for transfer')

                logger.info(f'Finished with config {configIndex + 1} of {config_paths_length}')

            status.reset('SUCCESS')

        if status.is_success() & (clearup is True):
            logger.info('Clearing up')

            clearup_dest_dir = f"{workspace_path}/clearup/{short_name}_{start_date}/"
            Path(clearup_dest_dir).mkdir(parents = True, exist_ok = True)

            logger.info(f"While developing, moving directories to this directory : {clearup_dest_dir}")

            clear_up(paths_to_clear, clearup_dest = clearup_dest_dir)

        end_script(premature = False)

    @abstractmethod
    def process_pre_job(self, args) -> bool:
        raise NotImplementedError

    @abstractmethod
    def process_in_job(self, jobPath, status, configjson, component) -> dict:
        raise NotImplementedError

    @abstractmethod
    def process_post_job(self, jobPath, configjson):
        raise NotImplementedError

    def run_processor(self, component: str, shortname: str):
        logger.info("Make sure to `conda activate py3EWSepi` environment!")
        logger.info("Make sure that flagdir package is available (on PYTHONPATH)")
        try:
            args: dict = self.parse_and_check_args()
            args["component"] = component
            args["short_name"] = shortname

            self.run_process(args)

        except SystemExit as e:
            print("caught with code " + str(e.code))
            logger.info('run_process() exited')
            sys.exit(e.code)
        except:
            raise Exception('Uncaught exception in run_Process:')
